"""Nanosurf nhf-file reader implementation for studio data
Copyright (C) Nanosurf AG - All Rights Reserved (2023)
License - MIT"""

import typing
import enum
import pathlib 
import re
import json
import numpy as np
import numpy.ma as ma
import h5py
from nanosurf.lib.datatypes import sci_val

class _CalMapType(enum.Enum):
    Linear                = enum.auto()
    Linear_Non_Invertible = enum.auto()
    Exponential           = enum.auto()
    Logarithmic           = enum.auto()
    Linear_Interpolation  = enum.auto()

class _CalNode():
    def __init__(self, uuid:str, sink_name:str, unit:str, map_type:_CalMapType, map_par:list[float], right_dir:bool=True, source_name:str= None):
        self.uuid = uuid
        self.sink_name = sink_name
        self.unit = unit
        self.source_name = source_name
        self.map_type = map_type
        self.map_par = map_par
        self.right_dir = right_dir

    def apply_calibration(self, source_data:np.ndarray) -> np.ndarray:
        if self.map_type == _CalMapType.Linear:
            a = self.map_par[0]
            b = self.map_par[1]
            if (a == 1.0) and (b == 0.0):
                return source_data
            else:
                if self.right_dir:
                    return source_data * a + b
                else:
                    return (source_data - b)/a
        elif self.map_type == _CalMapType.Linear_Non_Invertible:
            a = self.map_par[0]
            b = self.map_par[1]
            if self.right_dir:
                if (a == 1.0) and (b == 0.0):
                   return source_data
                else:
                   return source_data * a + b
            else:
                raise TypeError("Error: Unsupported calibration direction 'left_to_right' for Linear_Non_Invertible map type detected")    
        elif self.map_type == _CalMapType.Exponential:
            a = self.map_par[0]
            b = self.map_par[1]
            c = self.map_par[2]
            if self.right_dir:
                return b * np.exp((source_data-c)/a)
            else:
                return a * np.log(source_data/b) + c  
        elif self.map_type == _CalMapType.Logarithmic:
            a = self.map_par[0]
            b = self.map_par[1]
            c = self.map_par[2]
            if self.right_dir:
                return a * np.log(source_data/b) + c
            else:
                return b * np.exp((source_data-c)/a)  
        else:
            raise TypeError(f"Error: Unsupported calibration node type: {self.map_type=} detected")        

class NHFMeasurementType(enum.Enum):
    Undefined = enum.auto()
    Unknown = enum.auto()
    Image = enum.auto()
    Spectroscopy = enum.auto()
    WaveModeNMA = enum.auto()

def _default_verbose_output_handler(msg:str):
    print(msg)

def extract_unit(text_with_unit:str) -> str:
    def _isolate_unit(text_with_unit:str) -> str:
        """ Extracts the unit from a text string containing the unit in parentheses (e.g 'Meter (m)' return 'm')
            if no parentheses are found the returned string is empty.
        """
        si_unit = ""
        matches  = re.findall(r'\(([^\)]+)\)',text_with_unit)
        try:
            si_unit = matches[0]
        except IndexError:
            pass
        return si_unit      

    si_unit = _isolate_unit(text_with_unit)
    if si_unit == "":
        si_unit = text_with_unit
    return si_unit   

def get_attributes(instance: typing.Union[h5py.File,h5py.Group,h5py.Dataset]) -> dict[str]:
    """ Iterates over the attributes of the given instance and puts them to a dictionary.
    If the attribute for the data type is available, data type specific information is stored.

    Parameters
    ----------

        instance: h5py.File | h5py.Group | h5py.Dataset
            Instance within the .nhf file to be read.

    Return
    ------
        attributes_dict: dict
            Contains the attributes of the analyzed instance.

    """
    if not isinstance(instance, (h5py.File, h5py.Group, h5py.Dataset)):
        raise TypeError("Not supported type of instance provided")
    
    attributes_dict: dict = {}
    for attributes_key, attributes_val in instance.attrs.items():
        attributes_dict[attributes_key] = attributes_val
    return attributes_dict

def _get_sub_items_by_name(instance: typing.Union[h5py.File,h5py.Group], identifiers: list[str]) -> dict[str, str]:
    """ Reads the names of the sub items of the given instance and puts them to a dict, where key - segment id, value- segment name

    Parameters
    ----------

        instance: h5py.File | h5py.Group | h5py.Dataset
            Instance within the .nhf file to be read.
        identifiers: list of strings
            The attribute which identifies the items by its name

    Return
    ------
        item_names: dict
            Contains the names to the given instance.
            key - segment id, value- segment name

    """
    if isinstance(instance, (h5py.File, h5py.Group)):
        item_names: dict[str, str] = {}
        for seg_key in instance.keys():
            for id_name in identifiers:
                try:
                    item_names[seg_key] = instance[seg_key].attrs[id_name]
                    break
                except KeyError:
                    pass # try next name
        return item_names
    else:
        return None

def _get_unidentified_sub_groups(instance: typing.Union[h5py.File,h5py.Group], known_names: dict[str, str]) -> dict[str, h5py.Group]:
    """ Reads the names of the sub items of the given instance which are not already identified by the known_name dict and puts them into a dict
    """
    if isinstance(instance, (h5py.File, h5py.Group)):
        if known_names is None:
            known_names = {}
        item_names: dict[str, h5py.Group] = {}
        for seg_key in instance.keys():
            seg_key:str
            if seg_key not in known_names: 
                if isinstance(instance[seg_key],h5py.Group):
                    item_names[seg_key] = instance[seg_key]
                    sub_groups = _get_unidentified_sub_groups(item_names[seg_key],{})
                    if sub_groups is not None:
                        item_names[seg_key].groups = sub_groups
                    sub_dataset = _get_unidentified_sub_dataset(item_names[seg_key],{})
                    if sub_dataset is not None:
                        item_names[seg_key].datasets = sub_dataset
        return item_names
    else:
        return None 
            
def _get_unidentified_sub_dataset(instance: typing.Union[h5py.File,h5py.Group], known_names: dict[str, str]) -> dict[str, h5py.Group]:
    """ Reads the names of the datasets of the given instance which are not already identified by the known_name dict and puts them into a dict
    """
    if isinstance(instance, h5py.File) or isinstance(instance, h5py.Group):
        if known_names is None:
            known_names = {}
        item_names: dict[str, h5py.Group] = {}
        for seg_key in instance.keys():
            seg_key:str
            if seg_key not in known_names: 
                if isinstance(instance[seg_key],h5py.Dataset):
                    item_names[seg_key] = instance[seg_key]
        return item_names
    else:
        return None 
            

class NHFDataset():
    def __init__(self, parent:'NHFSegment', h5_dataset:h5py.Dataset, delay_load_of_data_from_file:bool=True) -> None:
        self.parent_segment = parent
        self.h5_dataset = h5_dataset
        self.attribute = get_attributes(self.h5_dataset)
        self._inject_data_conversion_attributes()
        if delay_load_of_data_from_file:
            # type hint is only correct after scale_dataset_to_physical_units() is called at data read time
            self.dataset:np.ndarray = self.h5_dataset
        else:
            self.h5_dataset = self._convert_h5py_dataset_to_numpy_array(self.h5_dataset)

    def has_nan_values(self) -> bool:
        nan_val = self.get_nan_value() 
        if np.isnan(nan_val): 
            # nan values cannot be looked up by simple 'val in array' method 
            has_nan = len(np.argwhere(np.isnan(self.dataset))) > 0
        else:
            has_nan = nan_val in self.dataset
        return has_nan

    def get_nan_value(self) -> float:
        nan_value = np.nan
        if 'target_nan_value' in self.attribute:
            nan_value = self.attribute['target_nan_value']
        return nan_value

    def get_masked_dataset(self, fill_value=None) -> ma.MaskedArray:
        if 'target_nan_value' in self.attribute:
            nan_value = self.attribute['target_nan_value']
            if np.isnan(nan_value): 
                # nan values cannot be found and replaced by simple masked_values() method 
                masked_dataset:ma.MaskedArray = ma.masked_array(data=self.dataset[:], mask=np.isnan(self.dataset), fill_value = np.finfo(np.float64).min, copy=True, shrink=True)  
            else:
                masked_dataset:ma.MaskedArray = ma.masked_values(self.dataset[:], value=nan_value, copy=True, shrink=True)  
        else:
            masked_dataset = ma.array(self.dataset, copy=True, shrink=True)  

        if fill_value is not None:
            masked_dataset.filled(fill_value)
            self.attribute['target_nan_value'] = fill_value        
        return masked_dataset
    
    def _convert_h5py_dataset_to_numpy_array(self,h5_dataset:h5py.Dataset) -> np.ndarray:
        if 'type_nan_value' in self.attribute:
            self.attribute['target_nan_value'] = self.attribute['type_nan_value'] 
        return np.array(h5_dataset, dtype=np.float64)
    
    def _need_conversion_to_physical_unit(self, target_signal_id:str=None) -> bool:
        """ Returns True if data conversion to target signal is not yet done. If target is omitted default signal is unsed. 
        """
        if target_signal_id is None:
            target_signal_id = self.get_default_signal()

        if extract_unit(self.attribute['signal_unit']) and extract_unit(self.attribute['signal_calibration_unit']) == 'int':
            return False
        else:
            if target_signal_id is not None:
                return target_signal_id != self.attribute["target_signal_id"]
            else:
                try:
                    _ = self.attribute['signal_calibration_min']
                    _ = self.attribute['signal_calibration_max']
                    _ = self.attribute['type_min']
                    _ = self.attribute['type_max']
                    return True
                except KeyError:
                    pass
        return False
    
    def _scale_with_calibration_attributes_of_dataset(self):
        try:
            signal_min = typing.cast(float,self.attribute['signal_calibration_min'])
            signal_max = typing.cast(float,self.attribute['signal_calibration_max'])
            type_min = typing.cast(float,self.attribute['type_min'])
            type_max = typing.cast(float,self.attribute['type_max'])
        except KeyError:
            signal_min = 0.0
            signal_max = 1.0
            type_min = 0.0
            type_max = 1.0
        
        try:
            calibration_factor = (signal_max - signal_min) / (type_max-type_min)
        except ZeroDivisionError:
            calibration_factor = 1.0
            signal_min = 0.0
            type_min = 0.0

        self.attribute['target_signal_unit'] = self.attribute['signal_calibration_unit']
        self.attribute['target_signal_id'] = self.attribute['signal_name']
        
        # prepare destination dataset with raw date
        self.dataset = self._convert_h5py_dataset_to_numpy_array(self.h5_dataset)

        # calculate scaling from dataset
        if self.has_nan_values():
            self.attribute['target_nan_value'] = self.attribute['type_nan_value'] 
            masked_dataset = self.get_masked_dataset()
            masked_dataset = typing.cast(ma.MaskedArray,(masked_dataset[:] - type_min) * calibration_factor + signal_min)
            nan_value = np.iinfo(np.int32).min if self.unit == 'int' else np.finfo(masked_dataset.dtype).min
            self.dataset = ma.filled(masked_dataset, fill_value=nan_value)
            self.attribute['target_nan_value'] = nan_value
        else:
            self.dataset = (self.dataset[:] - type_min) * calibration_factor + signal_min

    def get_default_signal(self) -> str:
        target_signal_id = None
        try:
            target_signal_id = self.attribute["signal_selected"]
            source_signal_id = self.attribute["signal_id"]
            if target_signal_id == source_signal_id:
                return None
        except KeyError:
            pass
        return target_signal_id
    
    def calc_dataset_by_calibration_graph(self, calibration_graph:list[int]):
        cal_nodes = self.parent_segment._file_hdl.cal_group.calibration_nodes

        self.attribute['target_signal_unit'] = cal_nodes[calibration_graph[-1]].unit
        self.attribute['target_signal_id']   = cal_nodes[calibration_graph[-1]].sink_name

        # prepare destination dataset with raw date
        self.dataset = self._convert_h5py_dataset_to_numpy_array(self.h5_dataset)

        if self.has_nan_values():
            masked_dataset = self.get_masked_dataset()

            # follow the calibration graph and update dataset
            for node_index in calibration_graph:
                masked_dataset = cal_nodes[node_index].apply_calibration(masked_dataset)

            nan_value = np.iinfo(np.int32).min if self.unit == 'int' else np.finfo(masked_dataset.dtype).min
            self.dataset = ma.filled(masked_dataset, fill_value=nan_value)
            self.attribute['target_nan_value'] = nan_value
        else:
            # follow the calibration graph and update dataset
            for node_index in calibration_graph:
                self.dataset = cal_nodes[node_index].apply_calibration(self.dataset)

    def _scale_with_calibration_graph(self, target_signal_id:str):
        calibration_graph = self._build_calibration_graph(target_signal_id)
        self.calc_dataset_by_calibration_graph(calibration_graph)
        
    def _build_calibration_graph(self, target_signal_id:str) -> list[int]:
        calibration_graph = []
        try:
            source_signal_id = self.attribute["signal_id"]
            calibration_graph = self.parent_segment._file_hdl.cal_group.build_calibration_graph(source_signal_id, target_signal_id)
        except KeyError as e:
            raise IOError(f"Missing complete signal calibration information for {self.attribute['name']}. Reason: {e}")
        return calibration_graph
    
    def _get_calib_node_by_index(self, node_index:int) -> _CalNode:
        return self.parent_segment._file_hdl.cal_group.calibration_nodes[node_index]

    def _find_calibration_node_by_uuid(self, uuid:str) -> _CalNode:
        nodes = self.parent_segment._file_hdl.cal_group.calibration_nodes
        for node in nodes:
            if node.uuid == uuid:
                return node
        return None
    
    def _find_calibration_node_by_sink_id(self, sink_id:str) -> _CalNode:
        nodes = self.parent_segment._file_hdl.cal_group.calibration_nodes
        for node in nodes:
            if node.sink_name == sink_id:
                return node
        return None
    
    def _find_calibration_node_by_unit(self, sink_unit) -> _CalNode:
        sink_unit = extract_unit(sink_unit)
        nodes = self.parent_segment._file_hdl.cal_group.calibration_nodes
        for node in nodes:
            if extract_unit(node.unit) == sink_unit:
                return node
        return None
    
    def _load_dataset_without_conversion(self, handle_nan:bool=False):
        self.dataset = self._convert_h5py_dataset_to_numpy_array(self.h5_dataset)

        self.attribute['target_signal_unit'] = self.attribute['signal_calibration_unit']
        self.attribute['target_signal_id'] = self.attribute['signal_name']

        if handle_nan and self.has_nan_values():
            masked_dataset = self.get_masked_dataset()
            nan_value = np.iinfo(np.int32).min if self.unit == 'int' else np.finfo(masked_dataset.dtype).min
            self.dataset = ma.filled(masked_dataset, fill_value=nan_value)
            self.attribute['target_nan_value'] = nan_value

        
    def _load_dataset_with_physical_unit(self, target_signal_id:str = None) -> None:
        """ If scaling information are provided in the Dataset, apply them.
        Converts the bit pattern of the saved raw data to the value of the given datatype saved in the attributes
        and scales it with the given calibration values.

        Attention: 
            'dataset' is read from file in this process, which can last long depending on array size.
            Also dataset is converted from h5py dataset into numpy.ndarray
        """
        if target_signal_id is None:
            target_signal_id = self.get_default_signal()

        if target_signal_id is None:
            self._scale_with_calibration_attributes_of_dataset()
        else:
            self._scale_with_calibration_graph(target_signal_id)

    def read_data(self, with_signal:str):
        if self._need_conversion_to_physical_unit(with_signal):
            self._load_dataset_with_physical_unit(with_signal)
        else:
            self._load_dataset_without_conversion(handle_nan=True)

        if self.unit == 'int':
            self.dataset = self.dataset.astype(dtype=np.int32)

    def signals(self) -> set[str]:
        cal_targets:set[str] = set()
        if "signal_calibration_source" in self.attribute:
            cal_node = self._find_calibration_node_by_uuid(self.attribute["signal_calibration_source"])
            cal_targets.add(cal_node.sink_name)
        elif "signal_calibration_source_0" in self.attribute:
            index = 0
            while (attr_node_name:=f"signal_calibration_source_{index}") in self.attribute:
                cal_node = self._find_calibration_node_by_uuid(self.attribute[attr_node_name])
                cal_targets.add(cal_node.sink_name)
                index += 1
        elif "signal_selected" in self.attribute:
            cal_targets.add(self.attribute["signal_selected"])
        elif "signal_id" in self.attribute:
            cal_targets.add(self.attribute["signal_id"])
        return cal_targets

    def units(self) -> set[str]:
        cal_targets_units:set[str] = set()
        if "signal_calibration_source" in self.attribute:
            cal_node = self._find_calibration_node_by_uuid(self.attribute["signal_calibration_source"])
            cal_targets_units.add(extract_unit(cal_node.unit))
        elif "signal_calibration_source_0" in self.attribute:
            index = 0
            while (attr_node_name:=f"signal_calibration_source_{index}") in self.attribute:
                cal_node = self._find_calibration_node_by_uuid(self.attribute[attr_node_name])
                cal_targets_units.add(extract_unit(cal_node.unit))
                index += 1
            return cal_targets_units
        elif "signal_selected" in self.attribute:
            cal_targets_units.add(extract_unit(self.attribute["signal_unit"]))
        elif "signal_id" in self.attribute:
            cal_targets_units.add(extract_unit(self.attribute["signal_unit"]))
        return cal_targets_units

    def get_signal_from_unit(self, unit:str) -> str:
        node = self._find_calibration_node_by_unit(extract_unit(unit))
        if node is not None:
            return node.sink_name
        return ""


    @property
    def unit(self) -> str:
        try:
            si_unit = extract_unit(self.attribute['target_signal_unit'])
        except KeyError:
            si_unit = ""
        return si_unit   

    @property
    def name(self) -> str:
        try:
            my_name = self.attribute['signal_name']
        except KeyError:
            my_name = ""
        return my_name

    def _inject_data_conversion_attributes(self):
        """ Search for attributes defining information about data value scaling 

        Parameters
        ----------
            attr_dict: dict
                dictionary to be analyzed and populate with conversion information.
        """
        if 'dataset_element_type' in self.attribute:
            element_type = self.attribute['dataset_element_type']
            if isinstance(element_type, str): # file version 2.x
                    try:
                        self.attribute['type_min']      = NHFFileReader.dataset_element_type_defined_as_str[element_type][0]
                        self.attribute['type_max']      = NHFFileReader.dataset_element_type_defined_as_str[element_type][1]
                        self.attribute['type']          = NHFFileReader.dataset_element_type_defined_as_str[element_type][2]
                        self.attribute['type_nan_value']= self.h5_dataset.fillvalue #NHFFileReader.dataset_element_type_defined_as_str[element_type][3]                         
                    except  IndexError:
                        raise IOError(f"Unknown dataset_element_type '{element_type}'. ")
                    
                    if 'signal_name' in self.attribute:
                        if 'signal_calibration_unit' not in self.attribute:
                            self.attribute['signal_calibration_unit'] = ""
                        if 'signal_calibration_max' not in self.attribute:
                            self.attribute['signal_calibration_max'] = 1.0
                        if 'signal_calibration_min' not in self.attribute:
                            self.attribute['signal_calibration_min'] = 0.0
            else: # file version 1.x
                try:
                    self.attribute['type_min'] = NHFFileReader.dataset_element_type_defined_as_int[element_type][0]
                    self.attribute['type_max'] = NHFFileReader.dataset_element_type_defined_as_int[element_type][1]
                    self.attribute['type']     = NHFFileReader.dataset_element_type_defined_as_int[element_type][2]
                except IndexError:
                    raise IOError(f"Unknown dataset_element_type number {element_type}. ")
                
                try:
                    if 'base_calibration_unit' in self.attribute:
                        self.attribute['signal_name'] = self.attribute['name']
                        self.attribute['signal_calibration_unit'] = self.attribute['base_calibration_unit']
                        self.attribute['signal_calibration_max'] = self.attribute['base_calibration_max']
                        self.attribute['signal_calibration_min'] = self.attribute['base_calibration_min']
                except KeyError:
                    raise IOError(f"Missing complete signal calibration information for {self.attribute['name']}")

        self.attribute['target_signal_unit'] = ""
        self.attribute['target_signal_id'] = ""
        
    def convert_to_matrix(self,size_x:int, size_y:int):
        self.dataset = np.flipud(np.reshape(np.array(self.dataset), (size_y, size_x)))

class NHFProperties():
    def __init__(self) -> None:
        pass

class NHFSegment():
    def __init__(self, name: str, file_hdl: 'NHFFileReader', hdf_group:h5py.Group) -> None:
        self._name = name
        self._file_hdl = file_hdl
        self._hdf_group = hdf_group
        self.channel: dict[str, NHFDataset] = {}
        self.group: dict[str, h5py.Group] = {}
        self.h5_datasets: dict[str, h5py.Dataset] = {}
        self.property:NHFProperties = NHFProperties()
        self.attribute = get_attributes(self._hdf_group) 
        self._dataset_size = (0,0)
        self._dataset_range = (0.0,0.0)
        self._inject_segment_configuration_properties(known_identifiers=['segment_configuration', 'scan_configuration'])

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def channels(self) -> list[NHFDataset]:
        return list(self.channel.values())
    
    @property
    def groups(self) -> list[h5py.Group]:
        return list(self.group.values())
    
    @property
    def datasets(self) -> list[h5py.Dataset]:
        return list(self.h5_datasets.values())
    
    @property
    def dataset_size_x(self) -> int:
        return self._dataset_size[0]

    @property
    def dataset_size_y(self) -> int:
        return self._dataset_size[1]

    @property
    def dataset_range_x(self) -> float:
        return self._dataset_range[0]

    @property
    def dataset_range_y(self) -> float:
        return self._dataset_range[1]

    def channel_signals(self, ch:typing.Union[str, int]) -> set[str]:
        nhf_dataset = None
        if isinstance(ch, str):
            nhf_dataset = self.channel[ch]
        elif isinstance(ch, int):
            nhf_dataset = self.channels[ch]
        else:
            raise TypeError(f"Parameter 'ch' has not supported data type of '{type(ch)}'")       
        return nhf_dataset.signals()        

    def channel_units(self, ch:typing.Union[str, int]) -> set[str]:
        nhf_dataset = None
        if isinstance(ch, str):
            nhf_dataset = self.channel[ch]
        elif isinstance(ch, int):
            nhf_dataset = self.channels[ch]
        else:
            raise TypeError(f"Parameter 'ch' has not supported data type of '{type(ch)}'")       
        return nhf_dataset.units()        

    def read_channel(self, ch:typing.Union[str, int], as_matrix:bool=False, with_signal:str=None, with_unit:str=None) -> NHFDataset:
        """ Loads a specified data channel by name or index.
            The data provided by physical units as specified in the files calibration information.

            For file version 2.0 and newer multiple calibrations per channel are possible. 
            The default calibration and unit is defined by the selected signal at the time of file storage.
            
            Returns:
            --------
                NHFDataset:
                    A class holding the channel data itself and supplementary information (e.g. unit)
                    If the channels has invalid numbers (e.g., not measured data points) these numbers are converted to NHFDataset.get_nan_value(). 
                    One can check if such numbers are in the data set by NHFDataset.has_nan_values(). 
                    if there are NaN values a masked data array can be provided by NHFDataSet.get_masked_dataset() to process only valid data points.

            Parameters:
            ------------
            ch: str, int
                Define the channel to load by its name or its index.

            as_matrix: bool, optional
                If set to True, the dataset is converted to a 2D numpy array. Otherwise, the data is a 1D array

            with_signal: str, optional
                If a signal name is provided, the corresponding calibration is used for the channel. 
                If not defined, the channels default signal is used.
                Possible signals can be accessed by self.channel_signals()

            with_unit: str
                If a unit name is provided, the corresponding calibration is used for the channel. 
                If not defined, the channels default unit is used.
                Possible signals can be accessed by self.channel_units()
        """
        nhf_dataset = None
        if isinstance(ch, str):
            nhf_dataset = self.channel[ch]
        elif isinstance(ch, int):
            nhf_dataset = self.channels[ch]
        else:
            raise TypeError(f"Parameter 'ch' has not supported data type of '{type(ch)}'")
        
        if (with_signal is not None) and self._file_hdl.version() < (2,0):
            raise ValueError("Parameter 'as_signal' is not supported by file version below 2.0")

        if (with_unit is not None) and self._file_hdl.version() < (2,0):
            raise ValueError("Parameter 'with_unit' is not supported by file version below 2.0")
        
        if with_unit is not None:
            known_units = nhf_dataset.units()
            if with_unit in known_units:
                with_signal = nhf_dataset.get_signal_from_unit(with_unit)
                if with_signal == "":
                    raise RuntimeError(f"Error: A target_id for known unit '{with_unit}' was not found.") 
            else:
                raise ValueError(f"Channel '{ch} does not contain calibration information for selected unit '{with_unit}'") 
                
        nhf_dataset.read_data(with_signal) 

        if as_matrix:
            if self._dataset_size == (0,0):
                raise ValueError("No size information stored in segment. Dataset cannot be converted to matrix.")
            nhf_dataset.convert_to_matrix(size_x=self.dataset_size_x, size_y=self.dataset_size_y)
        return nhf_dataset
    
    def channel_list(self) -> list[str]:
        return list(self.channel.keys())
    
    def channel_name(self, index: int) -> str:
        return self.channel_list()[index]
    
    def channel_count(self) -> int:
        return len(self.channel)
    
    def group_count(self) -> int:
        return len(self.group)        

    def group_list(self) -> list[str]:
        return list(self.group.keys())

    def group_name(self, key: int) -> str:
        return list(self.group.keys())[key]

    def dataset_count(self) -> int:
        return len(self.h5_datasets)
        
    def dataset_name(self, key: int) -> str:
        return list(self.h5_datasets.keys())[key]

    def find_dataset_by_attribute_value(self, attr_name:str, attr_value:typing.Any) -> h5py.Dataset:
        for dataset in self.h5_datasets.values():
            if attr_name in dataset.attrs:
                if dataset.attrs[attr_name] == attr_value:
                    return dataset
        return None
    
    def find_group_by_attribute_value(self, attr_name:str, attr_value:typing.Any) -> h5py.Group:
        for gr in self.group.values():
            if attr_name in gr.attrs:
                if gr.attrs[attr_name] == attr_value:
                    return gr
        return None
    
    def read_dataset_size(self) -> typing.Tuple[int, int]:
        rect_axis_size = (0,0)
        try:
            points_per_line = int(self.attribute['image_points_per_line'])
            number_of_lines = int(self.attribute['image_number_of_lines'])
            rect_axis_size = (points_per_line, number_of_lines)
        except KeyError:
            pass
        try:
            points_per_line = int(self.attribute["grid_number_of_columns"])
            number_of_lines = int(self.attribute["grid_number_of_rows"])
            rect_axis_size = (points_per_line, number_of_lines)   
        except KeyError:
            pass         
        try:
            points_per_line = int(self.attribute["rect_axis_size"][1])
            number_of_lines = int(self.attribute["rect_axis_size"][0])
            rect_axis_size = (points_per_line, number_of_lines)   
        except KeyError:
            pass         
        return rect_axis_size
    
    def read_dataset_range(self) -> typing.Tuple[float, float]:
        rect_axis_range = (0,0)
        try:
            range_x = float(self.attribute['image_size_x'])
            range_y = float(self.attribute['image_size_y'])
            rect_axis_range = (range_x, range_y)
        except KeyError:
            pass
        try:
            range_x = float(self.attribute["rect_axis_range"][1])
            range_y = float(self.attribute["rect_axis_range"][0])
            rect_axis_range = (range_x, range_y)   
        except KeyError:
            pass         
        try:
            range_x = self.attribute["grid_width"]
            range_y = self.attribute["grid_height"]
            rect_axis_range = (range_x, range_y)   
        except KeyError:
            pass         
        return rect_axis_range

    def _read_segment_structure(self):
        # read known dataset as 'channels'
        channel_identifier = 'signal_name' if self._file_hdl.version() >= (2,0) else 'name'        
        dataset_names = _get_sub_items_by_name(self._hdf_group, identifiers=[channel_identifier])
        if dataset_names:
            for dataset_id, dataset_name in dataset_names.items():
                dataset_data = self._hdf_group[dataset_id]
                self.channel[dataset_name] = NHFDataset(self,dataset_data)
        
        # read unknown dataset as 'group'
        self.group = _get_unidentified_sub_groups(self._hdf_group, dataset_names)
        self.h5_datasets = _get_unidentified_sub_dataset(self._hdf_group, dataset_names)
        self._dataset_size = self.read_dataset_size()
        self._dataset_range = self.read_dataset_range()
                
    def _inject_segment_configuration_properties(self, known_identifiers: list[str]):
        for identifier in known_identifiers:
            if identifier in  self.attribute:
                segment_config = json.loads(self.attribute[identifier])
                if 'property' in segment_config:
                    for prop_name, prop_value in segment_config['property'].items():
                            if isinstance(prop_value, dict) and 'value' in prop_value:
                                try: # to create a SciVal access to property
                                    # prepare value and unit
                                    prop_value['value'] = float(prop_value['value'])
                                    if 'unit' not in prop_value:
                                        prop_value['unit'] = ""

                                    # create attribute access to properties
                                    if prop_name not in self.attribute:
                                        self.attribute[prop_name] = prop_value
                                    else:
                                        self._file_hdl.print_verbose(f"Warning: Cannot create property '{prop_name}'. Attribute already exists")
    
                                    setattr(self.property,prop_name,sci_val.SciVal(value=prop_value['value'], unit_str=extract_unit(prop_value['unit'])))
                                except Exception:
                                    # properties seems not to be a number 
                                    if prop_name not in self.attribute:
                                        self.attribute[prop_name] = prop_value['value']
                                    else:
                                        self._file_hdl.print_verbose(f"Warning: Cannot create property '{prop_name}'. Attribute already exists")
                            else:
                                self._file_hdl.print_verbose(f"Warning: Unexpected property format of type '{type(prop_value)}' found")

class NHFMeasurement(NHFSegment):
    def __init__(self, name: str, file_hdl: 'NHFFileReader', hdf_group:h5py.Group) -> None:
        super().__init__(name, file_hdl, hdf_group)
        self.segment: dict[str, NHFSegment] = {}
        self._measurement_type = NHFMeasurementType.Undefined

    @property
    def segments(self) -> list[NHFSegment]:
        return list(self.segment.values())
    
    def segment_list(self) -> list[str]:
        return list(self.segment.keys())
    
    def segment_name(self, key: int) -> str:
        return self.segment_list()[key]
    
    def segment_count(self) -> int:
        return len(self.segment)
    
    @property
    def measurement_type(self) -> NHFMeasurementType:
        return self._detect_group_type()
    
    def read_dataset_size(self) -> typing.Tuple[int, int]:
        size = super().read_dataset_size()
        if (0,0) == size:
            for seg in self.segment.values():
                size = seg.read_dataset_size()
                if size != (0,0):
                    break
        return size
    
    def _detect_group_type(self) -> NHFMeasurementType:
        if self._measurement_type == NHFMeasurementType.Undefined:
            if self._file_hdl.version() >= (2,0):
                if 'group_type' in self.attribute:
                    current_group_type = self.attribute['group_type']
                    if current_group_type == 'image_line_based':
                        self._measurement_type = NHFMeasurementType.Image
                    elif current_group_type == 'spectroscopy':
                        self._measurement_type = NHFMeasurementType.Spectroscopy
                    elif current_group_type == 'wavemode_nma':
                        self._measurement_type = NHFMeasurementType.WaveModeNMA
                    else:
                        self._measurement_type = NHFMeasurementType.Unknown
            elif self._file_hdl.version() == (1,1):
                if 'measurement_type' in self.attribute:
                    current_group_type = self.attribute['measurement_type']
                    if current_group_type == 'image_line_based':
                        self._measurement_type = NHFMeasurementType.Image
                    elif current_group_type == 'spectroscopy_grid':
                        self._measurement_type = NHFMeasurementType.Spectroscopy
                    elif current_group_type == 'wavemode_nma':
                        self._measurement_type = NHFMeasurementType.WaveModeNMA
                    else:
                        self._measurement_type = NHFMeasurementType.Unknown

        if self._measurement_type == NHFMeasurementType.Unknown:
                self._file_hdl.print_verbose(f"Warning: Detecting group type for unknown file version: {self._file_hdl.version()}")
        if self._measurement_type == NHFMeasurementType.Undefined:
            self._file_hdl.print_verbose(f"Warning: measurement '{self.name}' has unknown 'group_type' = '{current_group_type}'")
        return self._measurement_type

    def _read_measurement_structure(self):
        segment_names = _get_sub_items_by_name(self._hdf_group, identifiers=['segment_name', 'name'])
        if segment_names:
            for segment_id, segment_name in segment_names.items():
                segment_data = self._hdf_group[segment_id]

                if isinstance(segment_data, h5py.Group):
                    self.segment[segment_name] = NHFSegment(segment_name, self._file_hdl, segment_data)
                    self.segment[segment_name]._read_segment_structure()
                elif isinstance(segment_data, h5py.Dataset):
                    self.channel[segment_name] = NHFDataset(self,segment_data)

        self.group = _get_unidentified_sub_groups(self._hdf_group, segment_names)
        self.h5_datasets = _get_unidentified_sub_dataset(self._hdf_group, segment_names)

        self._dataset_size = self.read_dataset_size()
        self._dataset_range = self.read_dataset_range()    
        for seg in self.segment.values():
            seg._dataset_range = self._dataset_range
            seg._dataset_size = self._dataset_size


class NHFCalibrationDataset():
    def __init__(self, file_hdl: 'NHFFileReader', hdf_dataset:h5py.Dataset) -> None:
        self._file_hdl = file_hdl
        self._hdf_dataset = hdf_dataset
        self.attribute = get_attributes(self._hdf_dataset) 

class NHFCalibrationGroup():
    def __init__(self, file_hdl: 'NHFFileReader', hdf_calibration_group:h5py.Group, group_id:str) -> None:
        self._file_hdl = file_hdl
        self._hdf_group = hdf_calibration_group
        self.group_id = group_id
        self.calibration_nodes:list[_CalNode] = []
        self.attribute = get_attributes(self._hdf_group) 
        self._load_calibration_graph()

    def build_calibration_graph(self, from__id_name:str, target_id_name:str) -> list[int]:
        calibration_graph:list[int] = []
        
        node_source_name = from__id_name
        exclude_sink_name = ""
        while (node_index:=self.find_calibration_node_by_name(node_source_name, exclude_sink_name)) >= 0:
            calib_node = self.calibration_nodes[node_index]
            calibration_graph.append(node_index)
            if calib_node.sink_name == target_id_name:
                return calibration_graph        
            node_source_name = calib_node.sink_name
            exclude_sink_name = calib_node.source_name
        return []

    def find_calibration_node_by_name(self, source_name:str, exclude_sink_name:str) -> int:
        for node_index, node in enumerate(self.calibration_nodes):
            if (node.source_name == source_name) and (node.sink_name != exclude_sink_name):
                    return node_index
        return -1

    def _load_calibration_graph(self) -> None:
        self.calibration_nodes.clear()
        for dataset in self._hdf_group.values():
            if not self._load_calibration_node(dataset):
                break

    def _load_calibration_node(self, hdf_dataset:h5py.Dataset) -> bool:
        done = False
        hdf_map_type_to_eunm = {
            "linear": _CalMapType.Linear,
            "linear_non_invertible": _CalMapType.Linear_Non_Invertible,
            "exponential": _CalMapType.Exponential,
            "logarithmic": _CalMapType.Logarithmic,
            "linear_interpolation": _CalMapType.Linear_Interpolation,
        }

        try:
            calib_attributes = get_attributes(hdf_dataset)
            signal_calibration_id = calib_attributes["signal_calibration_id"]
            signal_sink_id = calib_attributes["signal_sink_id"]
            signal_sink_unit = calib_attributes["signal_sink_unit"]
            signal_source_unit = calib_attributes["signal_source_unit"]
            signal_source_id = calib_attributes["signal_source_id"]
            signal_mapping_type = calib_attributes["signal_mapping_type"]
            signal_mapping_parameters = calib_attributes["signal_mapping_parameters"]
            signal_invertibility = calib_attributes["signal_invertibility"]

            # check attributes against know values
            try: 
                calib_map_type = hdf_map_type_to_eunm[signal_mapping_type]
            except KeyError:
                raise ValueError(f"Unsupported calibration attribute value: {signal_mapping_type=} in calibration dataset {signal_calibration_id=}")

            if signal_invertibility not in ["right","left", "two_way"]:
                raise ValueError(f"Unsupported calibration attribute value: {signal_invertibility=} in calibration dataset {signal_calibration_id=}")

            if signal_invertibility in ["right", "two_way"]:
                self.calibration_nodes.append(
                    _CalNode(
                        uuid = signal_calibration_id,
                        sink_name = signal_sink_id,
                        unit = signal_sink_unit,
                        map_type = calib_map_type,
                        map_par = signal_mapping_parameters,
                        right_dir = True,
                        source_name = signal_source_id
                    )
                )

            if signal_invertibility in ["left", "two_way"]:
                self.calibration_nodes.append(
                    _CalNode(
                        uuid = signal_calibration_id,
                        sink_name = signal_source_id,
                        unit = signal_source_unit,
                        map_type = calib_map_type,
                        map_par = signal_mapping_parameters,
                        right_dir = signal_invertibility == "left",
                        source_name = signal_sink_id
                    )
                )
            done = True

        except Exception as e:
            print(e)
        return done


class NHFFileReader():
    """ Main class to access nhf-files """

    # used by v1. NAN is not supported
    dataset_element_type_defined_as_int: dict = {
        0: [-(2.0**31), 2.0**31-1.0, 'dt_int32' ],
        1: [       0.0, 2.0**8 -1.0, 'dt_uint8' ],
        2: [       0.0, 2.0**16-1.0, 'dt_uint16'],
        3: [       0.0, 2.0**32-1.0, 'dt_uint32'],
        4: [       0.0, 2.0**64-1.0, 'dt_uint64'],
        5: [-(2.0**15), 2.0**15-1.0, 'dt_int16' ],
        6: [-(2.0**63), 2.0**63-1.0, 'dt_int64' ],
        7: [       0.0, 1.0        , 'dt_double']
    }
    # used by v2. NAN is defined as max "uint" or min "int"
    dataset_element_type_defined_as_str: dict = {
        'int32' : [-2.0**31 + 1.0, +2.0**31 - 1.0, 'dt_int32',  -2**31    ],
        'uint8' : [           0.0, +2.0**8  - 2.0, 'dt_uint8',   2**8  - 1],
        'uint16': [           0.0, +2.0**16 - 2.0, 'dt_uint16',  2**16 - 1],
        'uint32': [           0.0, +2.0**32 - 2.0, 'dt_uint32',  2**32 - 1],
        'uint64': [           0.0, +2.0**64 - 2.0, 'dt_uint64',  2**64 - 1],
        'int16' : [-2.0**15 + 1.0, +2.0**15 - 1.0, 'dt_int16' , -2**15    ],
        'int64' : [-2.0**63 + 1.0, +2.0**63 - 1.0, 'dt_int64' , -2**63    ],
        'double': [           0.0, +1.0          , 'dt_double',  np.nan]
    }
    
    def __init__(self, filename: pathlib.Path=None, verbose=False, verbose_handler=None):
        """ Provide a nhf-file path directly at creating of the class or call later read() with filename
         
        Parameters
        ----------
            verbose: bool
                Set this to True if messages during reading or accessing is desired
            
            verbose_handler: func(msg:str)
                Define an own message handler functions to redirect the messages 
                A None is provided the default message handler print the message to console
        """
        self.group: dict[str, h5py.Group] = {}        
        self.h5_datasets: dict[str, h5py.Dataset] = {}        
        self.measurement: dict[str, NHFMeasurement] = {}
        self.attribute: dict[str, typing.Any] = {}
        self._filename = filename
        self._file_id = None
        self._last_file_version = (0,0)
        self._verbose = verbose
        self._verbose_output_handler = verbose_handler if verbose_handler else _default_verbose_output_handler
        self._last_print_verbose_message = ""
        if self._filename:
            if not self.read():
                raise IOError(self._last_print_verbose_message)
            
    def version(self) -> typing.Tuple[int, int]:
        """ returns file version information in form of (major, minor) version number. If not accessible it returns (0,0)"""
        try: 
            return self._last_file_version
        except IndexError:
            return (0,0)
        
    def read(self, filename: pathlib.Path=None) -> bool:
        """ Open the nid-file with given path for read access. """
        self._clear_data()
        if filename is not None:
            self._filename = filename
        self._filename = pathlib.Path(self._filename)
        if self._filename.is_file():
            try:
                self._file_id = h5py.File(self._filename, 'r')
                self.attribute = get_attributes(self._file_id)
                self._read_version()
                file_version = self.version()
                if file_version == (1,1) or (file_version>=(2,0) and file_version<(3,0)):
                    self._read_file_structure()
                    return True
                else:
                    self.print_verbose(f"File version {self.version()} is not supported.")
                    return False
                    
            except Exception as e:
                if self._file_id is not None: 
                    self._file_id.close()
                self._file_id = None
                self._clear_data()
                self.print_verbose(f"Could not read structure of file.'\nReason: {e}")
                return False
        else:
           self.print_verbose(f"File does not exist: {self._filename}")
           return False
        

    @property
    def measurements(self) -> list[NHFMeasurement]:
        return list(self.measurement.values())
    
    def measurement_count(self) -> int:
        return len(self.measurement)
    
    def measurement_list(self) -> list[str]:
        return list(self.measurement.keys())
        
    def measurement_name(self, key: int) -> str:
        return list(self.measurement.keys())[key]
    
    @property
    def groups(self) -> list[h5py.Group]:
        return list(self.group.values())
        
    def group_count(self) -> int:
        return len(self.group)        

    def group_list(self) -> list[str]:
        return list(self.group.keys())

    def group_name(self, key: int) -> str:
        return self.group_list()[key]

    @property
    def datasets(self) -> list[h5py.Dataset]:
        return list(self.h5_datasets.values())
        
    def dataset_count(self) -> int:
        return len(self.h5_datasets)
        
    def dataset_list(self) -> list[str]:
        return list(self.h5_datasets.keys())

    def dataset_name(self, key: int) -> str:
        return self.dataset_list()[key]

    def print_verbose(self, msg:str):
        self._last_print_verbose_message = msg
        if self._verbose:
            self._verbose_output_handler(msg)

    def last_message(self) -> str:
        return self._last_print_verbose_message
    
    def pretty_print_structure(self):
        print(self)

    def print_file_structure(self):
        self.pretty_print_structure()

    # internal functions, not for user access
    
    def _read_version(self):
        major = int(self.attribute['nsf_file_version_major'])
        minor = int(self.attribute['nsf_file_version_minor'])
        self._last_file_version = (major, minor)        

    def _read_file_structure(self):    
        measurement_identifier = 'measurement_name' if self.version() >= (2,0) else 'name'
        measurement_names = _get_sub_items_by_name(self._file_id, identifiers=[measurement_identifier])
        if measurement_names:
            for measurement_id, measurement_name in measurement_names.items():
                measurement_data = self._file_id[measurement_id]
                self.measurement[measurement_name] = NHFMeasurement(measurement_name, self, measurement_data)
                self.measurement[measurement_name]._read_measurement_structure()

        self.cal_group = self._read_calibration_group()
        known_groups = measurement_names
        if self.cal_group is not None:
            known_groups[self.cal_group.group_id] = ""
        self.group = _get_unidentified_sub_groups(self._file_id, known_groups)
        self.h5_datasets = _get_unidentified_sub_dataset(self._file_id, known_groups)

    def _read_calibration_group(self) -> NHFCalibrationGroup:
        calib_group = None

        group_type_names = _get_sub_items_by_name(self._file_id, identifiers=["group_type"])  
        for group_name, group_type in  group_type_names.items():  
            if group_type == "calibration":
                calib_group = NHFCalibrationGroup(self, self._file_id[group_name], group_name)
                break
        return calib_group

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._clear_data()
        if self._file_id:
            self._file_id.close()
        self._file_id = None

    def __str__(self) -> str:
        structure = ""
        for m in self.measurements:
            structure += f"Measurement '{m.name}' of type '{m.measurement_type.name}':\n"
            for s in m.segments:
                structure += f"    Segment '{s.name}':\n"
                for ch in s.channels:
                    structure += f"        Channel '{ch.name}':\n"
                for gr in s.groups:
                        structure += f"        Group '{gr.name}':\n"
                for d in s.datasets:
                        structure += f"        Dataset '{d.name}':\n"
            for ch in m.channels:
                structure += f"    Channel '{ch.name}':\n"
            for gr in m.groups:
                structure += f"    Group '{gr.name}':\n"
            for d in m.datasets:
                structure += f"    Dataset '{d.name}':\n"
        if self.cal_group is not None:
            structure += f"Calibration Group '{self.cal_group.group_id}':\n"
        for gr in self.groups:
            structure += f"Group '{gr.name}':\n"
        for d in self.datasets:
            structure += f"Dataset '{d.name}':\n"
        return structure

    def __del__(self):
        if self._file_id:
            self._file_id.close()
        self._file_id = None

    def _clear_data(self):
        self.group: dict[str, h5py.Group] = {}
        self.measurement: dict[str, NHFMeasurement] = {}
        self.attribute: dict[str, typing.Any] = {}
